// Copyright 2019 Open Source Robotics Foundation, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/// @file
/// Trivial library to ensure we have some linking present for
/// `test_find_library`.

#include "rcpputils/visibility_control.hpp"

namespace test_library
{

RCPPUTILS_EXPORT
int add_one(int x)
{
  return x + 1;
}

}  // namespace test_library
