// Copyright 2020 Open Source Robotics Foundation, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef DUMMY_SHARED_LIBRARY__DUMMY_SHARED_LIBRARY_H_
#define DUMMY_SHARED_LIBRARY__DUMMY_SHARED_LIBRARY_H_

#if _WIN32
#ifdef DUMMY_SHARED_LIBRARY_BUILDING_DLL
#define DUMMY_SHARED_LIBRARY_PUBLIC __declspec(dllexport)
#else
#define DUMMY_SHARED_LIBRARY_PUBLIC __declspec(dllimport)
#endif
#else
#define DUMMY_SHARED_LIBRARY_PUBLIC
#endif

#include <stdio.h>

DUMMY_SHARED_LIBRARY_PUBLIC
void print_name(void);

#endif  // DUMMY_SHARED_LIBRARY__DUMMY_SHARED_LIBRARY_H_
